import json
import logging
from functools import lru_cache
from typing import Any, Optional
from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder

try:
    import redis  # type: ignore

    if int(redis.__version__.split(".", maxsplit=1)[0]) < 3:  # type: ignore
        raise Exception("Invalid version")
except Exception as err:
    raise Exception("Using jutil.redis_helpers requires redis>=3.0.0 installed") from err

logger = logging.getLogger(__name__)


@lru_cache()
def redis_connection_pool(connection_str: str) -> redis.ConnectionPool:
    """
    Returns cached Redis connection pool.
    :param connection_str: str
    :return: redis.ConnectionPool
    """
    return redis.ConnectionPool.from_url(connection_str)


@lru_cache()
def redis_instance(connection_str: str = "") -> redis.Redis:
    """
    Returns cached Redis instance using settings.REDIS_URL connection string.
    If no connection string is provided, AND settings.REDIS_URL is empty or missing, then
    localhost with default Redis settings (no password) is assumed.
    :param connection_str: Default connection string is settings.REDIS_URL or 'redis://:@localhost:6379/1?max_connections=50'.
    """
    if not connection_str:
        connection_str = settings.REDIS_URL if hasattr(settings, "REDIS_URL") else "redis://:@localhost:6379/1?max_connections=50"  # type: ignore
    return redis.Redis(connection_pool=redis_connection_pool(connection_str))


def redis_prefix_key(name: str) -> str:
    """
    Returns name prefixed with default DB name separated with dot.
    For example, if Django default DB name is "my_db"
    then "my_key" is returned as "my_db.my_key".
    :param name: Key name without prefix
    :return: Key name with prefix
    """
    return str(settings.DATABASES["default"]["NAME"]) + "." + name


def redis_set_bytes(name: str, value: bytes, ex: Optional[int] = None):
    """
    Sets value of the key as bytes to Redis.

    Uses default DB name as a key prefix. For example, if default DB name is "my_db"
    then key "my_key" is stored in Redis as "my_db.my_key".

    :param name: Key name without DB name prefix
    :param value: bytes
    :param ex: Optional expire time in seconds
    """
    return redis_instance().set(redis_prefix_key(name), value, ex=ex)


def redis_get_bytes_or_none(name: str) -> Optional[bytes]:
    """
    Returns value of the key as bytes from Redis or None if value missing.

    Uses default DB name as a key prefix. For example, if default DB name is "my_db"
    then key "my_key" is stored in Redis as "my_db.my_key".

    :param name: Key name without DB name prefix
    :return: bytes or None if value missing
    """
    return redis_instance().get(redis_prefix_key(name))


def redis_get_bytes(name: str) -> bytes:
    """
    Returns value of the key as bytes from Redis. Raise Exception if value is missing.

    Uses default DB name as a key prefix. For example, if default DB name is "my_db"
    then key "my_key" is stored in Redis as "my_db.my_key".

    :param name: Key name without DB name prefix
    :return: bytes
    """
    buf = redis_get_bytes_or_none(name)
    if buf is None:
        raise Exception(f"{redis_prefix_key(name)} not in Redis")
    return buf


def redis_set_json(name: str, value: Any, ex: Optional[int] = None, cls=DjangoJSONEncoder):
    """
    Sets value of the key as JSON to Redis.

    Uses default DB name as a key prefix. For example, if default DB name is "my_db"
    then key "my_key" is stored in Redis as "my_db.my_key".

    Stored value is assumed to be JSON and it is serialized before sending.

    :param name: Key name without DB name prefix
    :param value: Any value that will be serialized as JSON
    :param ex: Optional expire time in seconds
    :param cls: JSON encoder class (default is DjangoJSONEncoder from Django)
    """
    value_bytes = json.dumps(value, cls=cls).encode()
    return redis_instance().set(redis_prefix_key(name), value_bytes, ex=ex)


def redis_get_json_or_none(name: str) -> Any:
    """
    Returns value of the key as JSON from Redis or None if value missing.

    Uses default DB name as a key prefix. For example, if default DB name is "my_db"
    then key "my_key" is stored in Redis as "my_db.my_key".

    Stored value is assumed to be JSON and it is deserialized before returning.

    :param name: Key name without DB name prefix
    :return: bytes deserialized as JSON or None
    """
    try:
        buf = redis_instance().get(redis_prefix_key(name))
        if buf is not None:
            return json.loads(buf)
    except Exception as exc:
        logger.error("redis_get_json(%s) failed: %s", name, exc)
    return None


def redis_get_json(name: str) -> Any:
    """
    Returns value of the key as JSON from Redis. Raise Exception if value is missing.

    Uses default DB name as a key prefix. For example, if default DB name is "my_db"
    then key "my_key" is stored in Redis as "my_db.my_key".

    Stored value is assumed to be JSON and it is deserialized before returning.

    :param name: Key name without DB name prefix
    :return: bytes deserialized as JSON
    """
    buf = redis_instance().get(redis_prefix_key(name))
    if buf is None:
        raise Exception(f"{redis_prefix_key(name)} not in Redis")
    return json.loads(buf)


def redis_delete(name: str):
    """
    Deletes key-value pair from from Redis.

    Uses default DB name as a key prefix. For example, if default DB name is "my_db"
    then key "my_key" is stored in Redis as "my_db.my_key".

    :param name: Key name without DB name prefix
    """
    return redis_instance().delete(redis_prefix_key(name))



import logging
from urllib.parse import urlparse, parse_qsl, urlunparse, urlencode

logger = logging.getLogger(__name__)


def modify_url(url: str, new_params: dict) -> str:
    """
    Modifies existing URL by setting/overriding specified query string parameters.
    This can be useful for example if you need to modify user-provided callback URL
    for extra arguments. Note: Does not support multiple querystring parameters with identical name.
    :param url: Base URL/path to modify
    :param new_params: Querystring parameters to set/override (dict)
    :return: New URL/path
    """
    res = urlparse(url)
    query_params = dict(parse_qsl(res.query))
    for k, v in new_params.items():
        if v is None:
            query_params[str(k)] = ""
        else:
            query_params[str(k)] = str(v)
    parts = list(res)
    parts[4] = urlencode(query_params)
    return urlunparse(parts)



import os

from django.apps import apps
from django.conf import settings
from django.core.management.base import CommandParser
from jutil.command import SafeCommand


class Command(SafeCommand):
    help = "Lists project's top-level apps directly under BASE_DIR"

    def add_arguments(self, parser: CommandParser):
        parser.add_argument("--nl", action="store_true", help="Use \\n as app separator")

    def do(self, *args, **kw):
        sep = " "
        if kw["nl"]:
            sep = "\n"
        for app in apps.get_app_configs():
            name = app.name
            full_path = os.path.join(os.path.join(settings.BASE_DIR, name), "apps.py")
            if os.path.isfile(full_path):
                self.stdout.write(name, ending=sep)



from datetime import datetime, timedelta
from typing import Tuple, Any, Optional, List
import pytz
from calendar import monthrange
from django.utils.text import format_lazy
from django.utils.translation import gettext_lazy as _


TIME_RANGE_CHOICES = [
    ("yesterday", _("yesterday")),
    ("today", _("today")),
    ("tomorrow", _("tomorrow")),
    ("last_week", _("last week")),
    ("last_month", _("last month")),
    ("last_year", _("last year")),
    ("this_week", _("this week")),
    ("this_month", _("this month")),
    ("this_year", _("this year")),
    ("next_week", _("next week")),
    ("next_month", _("next month")),
    ("next_year", _("next year")),
]
# plus +- date ranges from current datetime:
# (e.g. --yesterday is full day yesterday but --prev-1d is 24h less from current time)
for d in [360, 180, 90, 60, 45, 30, 15, 7, 2, 1]:
    TIME_RANGE_CHOICES.extend(
        [
            ("prev_{}d".format(d), format_lazy("-{} {}", d, _("number.of.days"))),
            ("plus_minus_{}d".format(d), format_lazy("+-{} {}", d, _("number.of.days"))),
            ("next_{}d".format(d), format_lazy("+{} {}", d, _("number.of.days"))),
        ]
    )

TIME_RANGE_NAMES = list(zip(*TIME_RANGE_CHOICES))[0]

TIME_STEP_DAILY = "daily"
TIME_STEP_WEEKLY = "weekly"
TIME_STEP_MONTHLY = "monthly"

TIME_STEP_TYPES = [
    TIME_STEP_DAILY,
    TIME_STEP_WEEKLY,
    TIME_STEP_MONTHLY,
]

TIME_STEP_CHOICES = [
    (TIME_STEP_DAILY, _("daily")),
    (TIME_STEP_WEEKLY, _("weekly")),
    (TIME_STEP_MONTHLY, _("monthly")),
]

TIME_STEP_NAMES = list(zip(*TIME_STEP_CHOICES))[0]


def localize_time_range(begin: datetime, end: datetime, tz: Any = None) -> Tuple[datetime, datetime]:
    """
    Localizes time range. Uses pytz.utc if None provided.
    :param begin: Begin datetime
    :param end: End datetime
    :param tz: pytz timezone or None (default UTC)
    :return: begin, end
    """
    if tz is None:
        tz = pytz.utc
    return tz.localize(begin), tz.localize(end)


def get_last_day_of_month(today: Optional[datetime] = None) -> int:
    """
    Returns day number of the last day of the month
    :param today: Default UTC now
    :return: int
    """
    if today is None:
        today = datetime.utcnow()
    return monthrange(today.year, today.month)[1]


def end_of_month(today: Optional[datetime] = None, n: int = 0, tz: Any = None) -> datetime:
    """
    Returns end-of-month (last microsecond) of given datetime (or current datetime UTC if no parameter is passed).
    :param today: Some date in the month (defaults current datetime)
    :param n: +- number of months to offset from current month. Default 0.
    :param tz: Timezone (defaults pytz UTC)
    :return: datetime
    """
    if today is None:
        today = datetime.utcnow()
    last_day = monthrange(today.year, today.month)[1]
    end = today.replace(day=last_day, hour=0, minute=0, second=0, microsecond=0) + timedelta(hours=24)
    while n > 0:
        last_day = monthrange(end.year, end.month)[1]
        end = end.replace(day=last_day, hour=0, minute=0, second=0, microsecond=0) + timedelta(hours=24)
        n -= 1
    while n < 0:
        end -= timedelta(days=1)
        end = end.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        n += 1
    end_incl = end - timedelta(microseconds=1)
    if tz is None:
        tz = pytz.utc
    return tz.localize(end_incl)


def this_week(today: Optional[datetime] = None, tz: Any = None) -> Tuple[datetime, datetime]:
    """
    Returns this week begin (inclusive) and end (exclusive).
    Week is assumed to start from Monday (ISO).
    :param today: Some date (defaults current datetime)
    :param tz: Timezone (defaults pytz UTC)
    :return: begin (inclusive), end (exclusive)
    """
    if today is None:
        today = datetime.utcnow()
    begin = today - timedelta(days=today.weekday())
    begin = datetime(year=begin.year, month=begin.month, day=begin.day)
    return localize_time_range(begin, begin + timedelta(days=7), tz)


def this_month(today: Optional[datetime] = None, tz: Any = None) -> Tuple[datetime, datetime]:
    """
    Returns current month begin (inclusive) and end (exclusive).
    :param today: Some date in the month (defaults current datetime)
    :param tz: Timezone (defaults pytz UTC)
    :return: begin (inclusive), end (exclusive)
    """
    if today is None:
        today = datetime.utcnow()
    begin = datetime(day=1, month=today.month, year=today.year)
    end = begin + timedelta(days=32)
    end = datetime(day=1, month=end.month, year=end.year)
    return localize_time_range(begin, end, tz)


def this_year(today: Optional[datetime] = None, tz: Any = None) -> Tuple[datetime, datetime]:
    """
    Returns this year begin (inclusive) and end (exclusive).
    :param today: Some date (defaults current datetime)
    :param tz: Timezone (defaults pytz UTC)
    :return: begin (inclusive), end (exclusive)
    """
    if today is None:
        today = datetime.utcnow()
    begin = datetime(day=1, month=1, year=today.year)
    end = datetime(day=1, month=1, year=begin.year + 1)
    return localize_time_range(begin, end, tz)


def next_year(today: Optional[datetime] = None, tz: Any = None) -> Tuple[datetime, datetime]:
    if today is None:
        today = datetime.utcnow()
    begin = datetime(day=1, month=1, year=today.year + 1)
    end = datetime(day=1, month=1, year=begin.year + 2)
    return localize_time_range(begin, end, tz)


def next_week(today: Optional[datetime] = None, tz: Any = None) -> Tuple[datetime, datetime]:
    """
    Returns next week begin (inclusive) and end (exclusive).
    :param today: Some date (defaults current datetime)
    :param tz: Timezone (defaults pytz UTC)
    :return: begin (inclusive), end (exclusive)
    """
    if today is None:
        today = datetime.utcnow()
    begin = today + timedelta(days=7 - today.weekday())
    begin = datetime(year=begin.year, month=begin.month, day=begin.day)
    return localize_time_range(begin, begin + timedelta(days=7), tz)


def next_month(today: Optional[datetime] = None, tz: Any = None) -> Tuple[datetime, datetime]:
    """
    Returns next month begin (inclusive) and end (exclusive).
    :param today: Some date in the month (defaults current datetime)
    :param tz: Timezone (defaults pytz UTC)
    :return: begin (inclusive), end (exclusive)
    """
    if today is None:
        today = datetime.utcnow()
    begin = datetime(day=1, month=today.month, year=today.year)
    next_mo = begin + timedelta(days=32)
    begin = datetime(day=1, month=next_mo.month, year=next_mo.year)
    following_mo = begin + timedelta(days=32)
    end = datetime(day=1, month=following_mo.month, year=following_mo.year)
    return localize_time_range(begin, end, tz)


def last_week(today: Optional[datetime] = None, tz: Any = None) -> Tuple[datetime, datetime]:
    """
    Returns last week begin (inclusive) and end (exclusive).
    :param today: Some date (defaults current datetime)
    :param tz: Timezone (defaults pytz UTC)
    :return: begin (inclusive), end (exclusive)
    """
    if today is None:
        today = datetime.utcnow()
    begin = today - timedelta(weeks=1, days=today.weekday())
    begin = datetime(year=begin.year, month=begin.month, day=begin.day)
    return localize_time_range(begin, begin + timedelta(days=7), tz)


def last_month(today: Optional[datetime] = None, tz: Any = None) -> Tuple[datetime, datetime]:
    """
    Returns last month begin (inclusive) and end (exclusive).
    :param today: Some date (defaults current datetime)
    :param tz: Timezone (defaults pytz UTC)
    :return: begin (inclusive), end (exclusive)
    """
    if today is None:
        today = datetime.utcnow()
    end = datetime(day=1, month=today.month, year=today.year)
    end_incl = end - timedelta(seconds=1)
    begin = datetime(day=1, month=end_incl.month, year=end_incl.year)
    return localize_time_range(begin, end, tz)


def last_year(today: Optional[datetime] = None, tz: Any = None) -> Tuple[datetime, datetime]:
    """
    Returns last year begin (inclusive) and end (exclusive).
    :param today: Some date (defaults current datetime)
    :param tz: Timezone (defaults pytz UTC)
    :return: begin (inclusive), end (exclusive)
    """
    if today is None:
        today = datetime.utcnow()
    end = datetime(day=1, month=1, year=today.year)
    end_incl = end - timedelta(seconds=1)
    begin = datetime(day=1, month=1, year=end_incl.year)
    return localize_time_range(begin, end, tz)


def yesterday(today: Optional[datetime] = None, tz: Any = None) -> Tuple[datetime, datetime]:
    """
    Returns yesterday begin (inclusive) and end (exclusive).
    :param today: Some date (defaults current datetime)
    :param tz: Timezone (defaults pytz UTC)
    :return: begin (inclusive), end (exclusive)
    """
    if today is None:
        today = datetime.utcnow()
    end = datetime(day=today.day, month=today.month, year=today.year)
    end_incl = end - timedelta(seconds=1)
    begin = datetime(day=end_incl.day, month=end_incl.month, year=end_incl.year)
    return localize_time_range(begin, end, tz)


def add_month(t: datetime, n: int = 1) -> datetime:
    """
    Adds +- n months to datetime.
    Clamps days to number of days in given month.
    :param t: datetime
    :param n: +- number of months to offset from current month. Default 1.
    :return: datetime
    """
    t2 = t
    for count in range(abs(n)):  # pylint: disable=unused-variable
        if n > 0:
            t2 = datetime(year=t2.year, month=t2.month, day=1) + timedelta(days=32)
        else:
            t2 = datetime(year=t2.year, month=t2.month, day=1) - timedelta(days=2)
        try:
            t2 = t.replace(year=t2.year, month=t2.month)
        except Exception:
            last_day = monthrange(t2.year, t2.month)[1]
            t2 = t.replace(year=t2.year, month=t2.month, day=last_day)
    return t2


def per_delta(start: datetime, end: datetime, delta: timedelta):
    """
    Iterates over time range in steps specified in delta.

    :param start: Start of time range (inclusive)
    :param end: End of time range (exclusive)
    :param delta: Step interval

    :return: Iterable collection of [(start+td*0, start+td*1), (start+td*1, start+td*2), ..., end)
    """
    curr = start
    while curr < end:
        curr_end = curr + delta
        yield curr, curr_end
        curr = curr_end


def per_month(start: datetime, end: datetime, n: int = 1):
    """
    Iterates over time range in one month steps.
    Clamps to number of days in given month.

    :param start: Start of time range (inclusive)
    :param end: End of time range (exclusive)
    :param n: Number of months to step. Default is 1.

    :return: Iterable collection of [(month+0, month+1), (month+1, month+2), ..., end)
    """
    curr = start.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    while curr < end:
        curr_end = add_month(curr, n)
        yield curr, curr_end
        curr = curr_end


def get_time_steps(step_type: str, begin: datetime, end: datetime) -> List[Tuple[datetime, datetime]]:
    """
    Returns time stamps by time step type [TIME_STEP_DAILY, TIME_STEP_WEEKLY, TIME_STEP_MONTHLY].
    For example daily steps for a week returns 7 [begin, end) ranges for each day of the week.
    :param step_type: One of TIME_STEP_DAILY, TIME_STEP_WEEKLY, TIME_STEP_MONTHLY
    :param begin: datetime
    :param end: datetime
    :return: List of [begin, end), one for reach time step unit
    """
    after_end = end
    if TIME_STEP_DAILY == step_type:
        after_end += timedelta(days=1)
    elif TIME_STEP_WEEKLY == step_type:
        after_end += timedelta(days=7)
    elif TIME_STEP_MONTHLY == step_type:
        after_end = add_month(end)
    else:
        raise ValueError('Time step "{}" not one of {}'.format(step_type, TIME_STEP_TYPES))

    begins: List[datetime] = []
    t0 = t = begin
    n = 1
    while t < after_end:
        begins.append(t)
        if step_type == TIME_STEP_DAILY:
            t = t0 + timedelta(days=n)
        elif step_type == TIME_STEP_WEEKLY:
            t = t0 + timedelta(days=7 * n)
        elif step_type == TIME_STEP_MONTHLY:
            t = add_month(t0, n)
        n += 1
    return [(begins[i], begins[i + 1]) for i in range(len(begins) - 1)]



import logging
from typing import Optional
from uuid import uuid1
from django.contrib.auth import get_user_model
from django.contrib.auth.models import User
from rest_framework.test import APIClient
from rest_framework.authtoken.models import Token

logger = logging.getLogger(__name__)


class TestSetupMixin:
    @staticmethod
    def add_test_user(email: str = "", password: str = "", username: str = "", **kwargs) -> User:  # nosec
        """
        Add and login test user.
        :param email: Optional email. Default is <random>@example.com
        :param password: Optional password. Default is <random>.
        :param username: Optional username. Defaults to email.
        :return: User
        """
        if not email:
            email = "{}@example.com".format(username or uuid1().hex)
        if not password:
            email = email.split("@")[0]
        if not username:
            username = email
        user = get_user_model().objects.create(username=username, email=email, **kwargs)
        user.set_password(password)
        user.save()
        return user

    @staticmethod
    def create_api_client(user: Optional[User] = None) -> APIClient:
        """
        Creates APIClient with optionally authenticated user (by authorization token).
        :param user: User to authenticate (optional)
        :return: APIClient
        """
        token: Optional[Token] = None
        if user:
            token = Token.objects.get_or_create(user=user)[0]
            assert isinstance(token, Token)
        api_client = APIClient()
        if token:
            api_client.credentials(HTTP_AUTHORIZATION="Token {}".format(token.key))
        return api_client



import json
from collections import OrderedDict
from typing import Optional, Sequence, List, Dict, Any
from django.conf import settings
from django.contrib import admin
from django.contrib.auth import get_user_model
from django.contrib.auth.models import User
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import QuerySet
from django.http import HttpRequest
from django.urls import reverse, resolve
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from django.contrib.admin.models import CHANGE
from django.template.response import TemplateResponse
from django.contrib.admin.options import get_content_type_for_model
from django.contrib.admin.utils import unquote
from django.core.exceptions import PermissionDenied
from django.utils.text import capfirst
from django.utils.encoding import force_str
from django.contrib.admin.models import LogEntry


def get_admin_log(instance: object) -> QuerySet:
    """
    Returns admin log (LogEntry QuerySet) of the object.
    """
    return LogEntry.objects.filter(
        content_type_id=get_content_type_for_model(instance).pk,  # type: ignore
        object_id=instance.pk,  # type: ignore  # pytype: disable=attribute-error
    )


def admin_log(instances: Sequence[object], msg: str, who: Optional[User] = None, **kwargs):
    """
    Logs an entry to admin logs of model(s).
    :param instances: Model instance or list of instances (None values are ignored)
    :param msg: Message to log
    :param who: Who did the change. If who is None then User with username of settings.DJANGO_SYSTEM_USER (default: 'system') will be used
    :param kwargs: Optional key-value attributes to append to message
    :return: None
    """
    # use system user if 'who' is missing
    if who is None:
        username = settings.DJANGO_SYSTEM_USER if hasattr(settings, "DJANGO_SYSTEM_USER") else "system"  # type: ignore
        who = get_user_model().objects.get_or_create(username=username)[0]

    # allow passing individual instance
    if not isinstance(instances, list) and not isinstance(instances, tuple):
        instances = [instances]  # type: ignore

    # append extra context if any
    cx: Dict[str, Any] = {}
    for k, v in kwargs.items():
        if hasattr(v, "pk"):  # log only primary key for model instances, not whole str representation
            v = v.pk
        cx[k] = v
    if cx:
        msg += " | " + json.dumps(cx, cls=DjangoJSONEncoder)

    for instance in instances:
        if instance:
            LogEntry.objects.log_action(
                user_id=who.pk if who is not None else None,
                content_type_id=get_content_type_for_model(instance).pk,  # type: ignore
                object_id=instance.pk,  # type: ignore  # pytype: disable=attribute-error
                object_repr=force_str(instance),
                action_flag=CHANGE,
                change_message=msg,
            )


def admin_log_changed_fields(obj: object, field_names: Sequence[str], who: Optional[User] = None, **kwargs):
    """
    Logs changed fields of a model instance to admin log.
    :param obj: Model instance
    :param field_names: Field names
    :param who: Who did the change. If who is None then User with username of settings.DJANGO_SYSTEM_USER (default: 'system') will be used
    :param kwargs: Optional key-value attributes to append to message
    :return:
    """
    from jutil.model import get_model_field_label_and_value  # noqa

    fv: List[str] = []
    for k in field_names:
        label, value = get_model_field_label_and_value(obj, k)
        fv.append('{}: "{}"'.format(label, value))
    msg = ", ".join(fv)
    if "ip" in kwargs:
        msg += " (IP {ip})".format(ip=kwargs.pop("ip"))
    admin_log([obj], msg, who, **kwargs)  # type: ignore


def admin_obj_url(obj: Optional[object], route: str = "", base_url: str = "") -> str:
    """
    Returns admin URL to object. If object is standard model with default route name, the function
    can deduct the route name as in "admin:<app>_<class-lowercase>_change".
    :param obj: Object
    :param route: Empty for default route
    :param base_url: Base URL if you want absolute URLs, e.g. https://example.com
    :return: URL to admin object change view
    """
    if obj is None:
        return ""
    if not route:
        route = "admin:{}_{}_change".format(obj._meta.app_label, obj._meta.model_name)  # type: ignore
    path = reverse(route, args=[obj.id])  # type: ignore
    return base_url + path


def admin_obj_link(obj: Optional[object], label: str = "", route: str = "", base_url: str = "") -> str:
    """
    Returns safe-marked admin link to object. If object is standard model with default route name, the function
    can deduct the route name as in "admin:<app>_<class-lowercase>_change".
    :param obj: Object
    :param label: Optional label. If empty uses str(obj)
    :param route: Empty for default route
    :param base_url: Base URL if you want absolute URLs, e.g. https://example.com
    :return: HTML link marked safe
    """
    if obj is None:
        return ""
    url = mark_safe(admin_obj_url(obj, route, base_url))  # nosec
    return format_html("<a href='{}'>{}</a>", url, str(obj) if not label else label)


class ModelAdminBase(admin.ModelAdmin):
    """
    ModelAdmin with save-on-top default enabled and customized (length-limited) history view.
    """

    save_on_top = True
    max_history_length = 1000

    def sort_actions_by_description(self, actions: dict) -> OrderedDict:
        """
        :param actions: dict of str: (callable, name, description)
        :return: OrderedDict
        """
        sorted_descriptions = sorted([(k, data[2]) for k, data in actions.items()], key=lambda x: x[1])
        sorted_actions = OrderedDict()
        for k, description in sorted_descriptions:  # pylint: disable=unused-variable
            sorted_actions[k] = actions[k]
        return sorted_actions

    def get_actions(self, request):
        return self.sort_actions_by_description(super().get_actions(request))

    def fill_extra_context(self, request: HttpRequest, extra_context: Optional[Dict[str, Any]]):  # pylint: disable=unused-argument
        """
        Function called by customized add_view(), change_view() and kw_changelist_view()
        to supplement extra_context dictionary by custom variables.
        """
        return extra_context

    def add_view(self, request: HttpRequest, form_url="", extra_context=None):
        """
        Custom add_view() which calls fill_extra_context().
        """
        return super().add_view(request, form_url, self.fill_extra_context(request, extra_context))

    def change_view(self, request: HttpRequest, object_id, form_url="", extra_context=None):
        """
        Custom change_view() which calls fill_extra_context().
        """
        return super().change_view(request, object_id, form_url, self.fill_extra_context(request, extra_context))

    def kw_changelist_view(self, request: HttpRequest, extra_context=None, **kwargs):  # pylint: disable=unused-argument
        """
        Changelist view which allow key-value arguments and calls fill_extra_context().
        :param request: HttpRequest
        :param extra_context: Extra context dict
        :param kwargs: Key-value dict
        :return: See changelist_view()
        """
        return self.changelist_view(request, self.fill_extra_context(request, extra_context))

    def history_view(self, request, object_id, extra_context=None):
        """
        The 'history' admin view for this model.
        """
        from django.contrib.admin.models import LogEntry  # noqa

        # First check if the user can see this history.
        model = self.model
        obj = self.get_object(request, unquote(object_id))
        if obj is None:
            return self._get_obj_does_not_exist_redirect(request, model._meta, object_id)

        if not self.has_view_or_change_permission(request, obj):
            raise PermissionDenied

        # Then get the history for this object.
        opts = model._meta
        app_label = opts.app_label
        action_list = (
            LogEntry.objects.filter(object_id=unquote(object_id), content_type=get_content_type_for_model(model))
            .select_related()
            .order_by("-action_time")[: self.max_history_length]
        )

        context = {
            **self.admin_site.each_context(request),
            "title": _("Change history: %s") % obj,
            "action_list": action_list,
            "module_name": str(capfirst(opts.verbose_name_plural)),
            "object": obj,
            "opts": opts,
            "preserved_filters": self.get_preserved_filters(request),
            **(extra_context or {}),
        }

        request.current_app = self.admin_site.name

        return TemplateResponse(
            request,
            self.object_history_template
            or [
                "admin/%s/%s/object_history.html" % (app_label, opts.model_name),
                "admin/%s/object_history.html" % app_label,
                "admin/object_history.html",
            ],
            context,
        )


class InlineModelAdminParentAccessMixin:
    """
    Admin mixin for accessing parent objects to be used in InlineModelAdmin derived classes.
    """

    OBJECT_PK_KWARGS = ["object_id", "pk", "id"]

    def get_parent_object(self, request) -> Optional[object]:
        """
        Returns the inline admin object's parent object or None if not found.
        """
        mgr = self.parent_model.objects  # type: ignore
        resolved = resolve(request.path_info)
        if resolved.kwargs:
            for k in self.OBJECT_PK_KWARGS:
                if k in resolved.kwargs:
                    return mgr.filter(pk=resolved.kwargs[k]).first()
        if resolved.args:
            return mgr.filter(pk=resolved.args[0]).first()
        return None


class AdminLogEntryMixin:
    """
    Model mixin for logging Django admin changes of Models.
    Call fields_changed() on change events.
    """

    def fields_changed(self, field_names: Sequence[str], who: Optional[User] = None, **kwargs):
        admin_log_changed_fields(self, field_names, who, **kwargs)
